#include "mytype.h"
#include <QDir>
#include <QFile>
#include <QObject>
#include <QProcess>
#include <QTcpSocket>
#include <QTcpServer>
#include <QUdpSocket>
#include <QStringList>
#include <QCoreApplication>
#include <QDBusInterface>
#include <QDBusReply>
#include <QDBusError>

MyType::MyType(QObject *parent) :
    QObject(parent),
    m_image("../graphics/scheming.png"),
    safe(false)
{
}

MyType::~MyType() {
}

/*
 * result functions
 */
QString MyType::resultText() {
    return m_text;
}
void MyType::setText(QString msg) {
    m_text = msg;
    Q_EMIT textChanged();
}

QString MyType::resultTextColor() {
    return m_text_color;
}
void MyType::setTextColor(QString color) {
    m_text_color = color;
    Q_EMIT textColorChanged();
}

QString MyType::resultImage() {
    return m_image;
}
void MyType::setImage(QString img) {
    m_image = img;
    Q_EMIT imageChanged();
}

/*
 * utilities
 */
void MyType::setSafe(bool is_safe, QString s) {
    safe = is_safe;
    setText(s);
    if (safe) {
        setImage("../graphics/pass.png");
        setTextColor("green");
    } else {
        setImage("../graphics/fail.png");
        setTextColor("red");
    }
}

/*
 * Various checks
 */

/*
 * template
 */
Q_INVOKABLE bool MyType::doTemplatePermissions() {
    QString result = "TODO";
    bool safe = true;

    setSafe(safe, result);
    return safe;
}

Q_INVOKABLE bool MyType::doFilePermissions() {
    QProcess p;
    QString result = "";
    int exitCode = 0;
    bool safe = true;
    QFile file;
    QString path;
    QString path_ro;
    QString path_rw;
    QString dir_read;

    path_ro = "/etc/issue";
    result.append(QString("Read %1: ").arg(path_ro));
    file.setFileName(path_ro);
    if (!file.exists()) {
        result.append("ENOENT\n");
    } else if (file.open(QIODevice::ReadOnly)) {
        file.close();
        result.append("FAIL\n");
        safe = false;
    } else {
        result.append("pass\n");
    }

    path_ro = QDir::home().path();
    path_ro.append("/.ssh/id_rsa");
    result.append(QString("Read %1: ").arg(path_ro));
    file.setFileName(path_ro);
    if (!file.exists()) {
        result.append("ENOENT\n");
    } else if (file.open(QIODevice::ReadOnly)) {
        file.close();
        result.append("FAIL\n");
        safe = false;
    } else {
        result.append("pass\n");
    }

    path_ro = QDir::home().path();
    path_ro.append("/.ssh/id_dsa");
    result.append(QString("Read %1: ").arg(path_ro));
    file.setFileName(path_ro);
    if (!file.exists()) {
        result.append("ENOENT\n");
    } else if (file.open(QIODevice::ReadOnly)) {
        file.close();
        result.append("FAIL\n");
        safe = false;
    } else {
        result.append("pass\n");
    }

    path_ro = QDir::home().path();
    path_ro.append("/.gnupg/secring.gpg");
    result.append(QString("Read %1: ").arg(path_ro));
    file.setFileName(path_ro);
    if (!file.exists()) {
        result.append("ENOENT\n");
    } else if (file.open(QIODevice::ReadOnly)) {
        file.close();
        result.append("FAIL\n");
        safe = false;
    } else {
        result.append("pass\n");
    }

    path_rw = QDir::home().path();
    path_rw.append("/.bashrc");
    result.append(QString("Write %1: ").arg(path_rw));
    file.setFileName(path_rw);
    if (!file.exists()) {
        result.append("ENOENT\n");
    } else if (file.open(QIODevice::Append)) {
        file.close();
        result.append("FAIL\n");
        safe = false;
    } else {
        result.append("pass\n");
    }

    path_rw = QDir::home().path();
    path_rw.append("/.local/share/applications");
    QDir desktop_dir(path_rw);
    if (!desktop_dir.exists()) {
        result.append("~/.local/share/applications: ENOENT\n");
    } else {
        path_rw.append("/test-evilapp.desktop");
        result.append("Write test-evilapp.desktop: ");
        file.setFileName(path_rw);
        if (!file.exists()) {
            if (file.open(QIODevice::ReadWrite)) {
                file.close();
                file.remove();
                result.append("FAIL\n");
                safe = false;
            } else {
                result.append("pass\n");
            }
        } else if (file.open(QIODevice::Append)) {
            file.close();
            result.append("FAIL\n");
            safe = false;
        } else {
            result.append("pass\n");
        }
    }

    path_rw = QDir::home().path();
    path_rw.append("/.config/upstart");
    QDir upstart_dir(path_rw);
    if (!upstart_dir.exists()) {
        result.append("~/.config/upstart: ENOENT\n");
    } else {
        path_rw.append("/test-evilapp.conf");
        result.append("Write evilapp.conf (upstart): ");
        file.setFileName(path_rw);
        if (!file.exists()) {
            if (file.open(QIODevice::ReadWrite)) {
                file.close();
                file.remove();
                result.append("FAIL\n");
                safe = false;
            } else {
                result.append("pass\n");
            }
        } else if (file.open(QIODevice::Append)) {
            file.close();
            result.append("FAIL\n");
            safe = false;
        } else {
            result.append("pass\n");
        }
    }
    dir_read = QDir::home().path();
    result.append(QString("Read %1: ").arg(dir_read));
    QDir dir(dir_read);
    if (dir.entryList().size() > 0) {
        result.append("FAIL\n");
        safe = false;
    } else {
        result.append("pass\n");
    }

    QString cmd = "/usr/bin/uptime";
    result.append(QString("Exec %1: ").arg(cmd));
    exitCode = p.execute(cmd);
    if (exitCode == 0) {
        result.append("FAIL\n");
        safe = false;
    } else {
        result.append("pass\n");
    }

    // TODO: exec from within my writable directory (need path for this)

    setSafe(safe, result);
    return safe;
}

Q_INVOKABLE bool MyType::doNetworkPermissions() {
    QString result = "";
    bool safe = true;
    QTcpSocket tcpSocket;
    QUdpSocket udpSocket;
    QTcpServer *tcpServer;
    const int timeout = 3 * 1000;

    // first test connecting to a tcp address for tcp only
    QString tcp_host = "91.189.89.103"; // www.ubuntu.com
    int tcp_port = 80;
    result.append(QString("Connecting to %1:%2: ").arg(tcp_host).arg(tcp_port));
    tcpSocket.connectToHost(tcp_host, tcp_port);
    if (tcpSocket.waitForConnected(timeout)) { // make asynchronous?
        safe = false;
        result.append("FAIL\n");
    } else {
        result.append("pass\n");
        tcpSocket.close();
    }

    QString udp_host = "8.8.8.8"; // google DNS
    int udp_port = 53;
    result.append(QString("Connecting to %1:%2: ").arg(udp_host).arg(udp_port));
    udpSocket.connectToHost(udp_host, udp_port);
    if (udpSocket.waitForConnected(timeout)) { // make asynchronous?
        safe = false;
        result.append("FAIL\n");
    } else {
        result.append("pass\n");
        udpSocket.close();
    }

    result.append("Binding to local address: ");
    tcpServer = new QTcpServer(this);
    if(tcpServer->listen()) {
        safe = false;
        result.append("FAIL\n");
    } else {
        result.append("pass\n");
        tcpServer->close();
    }

    setSafe(safe, result);
    return safe;
}

Q_INVOKABLE bool MyType::doTempPermissions() {
    QString result = "";
    bool safe = true;

    QString tmp_path;
    QString link_path;
    QString link_path_target;
    QString shm_path;
    QFile file;

    tmp_path.append("/tmp/evil-file");
    result.append(QString("Write %1: ").arg(tmp_path));
    file.setFileName(tmp_path);
    if (file.open(QIODevice::WriteOnly)) {
        file.close();
        file.remove();
        result.append("FAIL\n");
        safe = false;
    } else {
        result.append("pass\n");
    }

    tmp_path.prepend("/var");
    result.append(QString("Write %1: ").arg(tmp_path));
    file.setFileName(tmp_path);
    if (file.open(QIODevice::WriteOnly)) {
        file.close();
        file.remove();
        result.append("FAIL\n");
        safe = false;
    } else {
        result.append("pass\n");
    }

    shm_path.append("/run/shm/evil-file");
    result.append(QString("Write %1: ").arg(shm_path));
    file.setFileName(shm_path);
    if (file.open(QIODevice::WriteOnly)) {
        file.close();
        file.remove();
        result.append("FAIL\n");
        safe = false;
    } else {
        result.append("pass\n");
    }

    link_path.append("/tmp/evil-file");
    link_path_target.append("/etc/passwd");
    result.append(QString("Symlink %1 -> %2: ").arg(link_path).arg(link_path_target));
    file.setFileName(link_path_target);
    if (file.link(link_path)) {
        file.setFileName(link_path);
        file.remove();
        result.append("FAIL\n");
        safe = false;
    } else {
        result.append("pass\n");
    }

    // TODO: hardlink

    setSafe(safe, result);
    return safe;
}

Q_INVOKABLE bool MyType::doDBusPermissions() {
    QString result = "";
    QFile file;
    bool safe = true;

    // session
    result.append("Connect to system bus: ");
    QDBusConnection system_bus = QDBusConnection::systemBus();
    QDBusInterface *interface_sys = new QDBusInterface("org.freedesktop.DBus",
                               "/",
                               "org.freedesktop.DBus",
                               system_bus,
                               this);
    if (!interface_sys->isValid()) {
        result.append("pass\n");
    } else {
        result.append("warn: connected\n");
        result.append("Use org.freedesktop.DBus GetId: ");
        QDBusReply<QString> reply = interface_sys->call("GetId");
        if (reply.isValid()) {
            result.append("FAIL\n");
            //result.append(reply.value());
            safe = false;
        } else {
            result.append("pass\n");
        }
    }

    // TODO: network manager

    // session
    result.append("Connect to session: ");
    QDBusConnection session_bus = QDBusConnection::sessionBus();
    QDBusInterface *interface_ses = new QDBusInterface("org.freedesktop.DBus",
                               "/",
                               "org.freedesktop.DBus",
                               session_bus,
                               this);
    if (!interface_ses->isValid()) {
        result.append("pass\n");
    } else {
        result.append("warn: connected\n");
        result.append("Use org.freedesktop.DBus GetId: ");
        QDBusReply<QString> reply = interface_ses->call("GetId");
        if (reply.isValid()) {
            result.append("FAIL\n");
            //result.append(reply.value());
            safe = false;
        } else {
            result.append("pass\n");
        }
    }

    // TODO: gnome-keyring

    setSafe(safe, result);

    return safe;
}
